from gentleagents.engines.openai_client import interact
from gentleagents.validations.models import ToolResponse
from pydantic import ValidationError
from typing import Callable, Dict, Any, get_type_hints
import json
from gentleagents.agents.helper import get_function_parameters, format_final_response
from gentleagents.queries.query import get_agent_summary

class Agent:
    def __init__(self, name, role, tools=None, model = None):
        """
        Initialize an AI Agent.

        Args:
            name (str): The agent's name.
            role (str): The agent's role (helps define its behavior).
            tools (list, optional): Functions the agent can use (default: None).
            model (str): The model to use as the engine
        """
        self.name = name
        self.role = role
        self.tools = tools or []
        self.model = model
        self.chat_history = []  

    def start_agent(self, user_message: str) -> str:
        try:
            self.chat_history.append({"role": "user", "content": user_message})

            tools_spec = []
            for name, func in self.tools.items():
                try:
                    tools_spec.append({
                        "type": "function",
                        "function": {
                            "name": name,
                            "description": func.__doc__ or "No description provided.",
                            "parameters": get_function_parameters(func)
                        }
                    })
                except Exception as e:
                    print(f"Error processing tool '{name}': {e}")
                    continue  

            try:
                response = interact(
                    system_prompt=f"You are {self.name}, a {self.role}.",
                    user_message=user_message,
                    tools=tools_spec,
                    model=self.model
                )
            except Exception as e:
                return f"Error communicating with AI model: {e}"

            message = response.choices[0].message
            self.chat_history.append({"role": "assistant", "content": message.content if message.content else ""})

            if not message.tool_calls:
                return message.content  

            responses = []
            for tool_call in message.tool_calls:
                function_name = tool_call.function.name
                function_args = tool_call.function.arguments

                if function_name not in self.tools:
                    responses.append(f"Error: Unknown tool '{function_name}'")
                    continue

                try:
                    if isinstance(function_args, str):
                        function_args = json.loads(function_args)

                    if not isinstance(function_args, dict):
                        raise TypeError(f"Expected dictionary for function arguments, got {type(function_args)}")

                    result = self.tools[function_name](**function_args)
                    type_hints = get_type_hints(self.tools[function_name])
                    return_type = type_hints.get("return", type(None))
                    validated_response = ToolResponse[return_type](result=result, message="Task completed with tool successfully.")
                    responses.append(f"Tool '{function_name}' executed successfully: {validated_response}")

                except ValidationError as e:
                    responses.append(f"Validation error in tool '{function_name}': {e}")
                except TypeError as e:
                    responses.append(f"Type error in tool '{function_name}': {e}")
                except Exception as e:
                    responses.append(f"Unexpected error during execution of '{function_name}': {e}")

            tool_responses = "\n".join(responses)
            self.chat_history.append({"role": "assistant", "content": tool_responses})

            summary_prompt = get_agent_summary(self.chat_history)

            try:
                final_response = interact(
                    system_prompt=f"You are {self.name}, a {self.role}.",
                    user_message=summary_prompt,
                    tools=None,
                    model=self.model
                )
            except Exception as e:
                return f"Error communicating with AI model for final response: {e}"

            final_message = final_response.choices[0].message.content
            self.chat_history.append({"role": "assistant", "content": final_message})

            print(f"{format_final_response(tool_responses)}\n\n{final_message}")
            return f"{format_final_response(tool_responses)}\n\n{final_message}"
        except Exception as e:
            return f"Critical error in agent execution: {e}"




from pydantic import BaseModel, Field
from typing import TypeVar, Generic, Any, List

T = TypeVar('T')

class ToolResponse(BaseModel, Generic[T]):  
    result: T
    message: str = Field(..., description="Message describing the result of the tool execution")

class TaskAssignment(BaseModel):
    agent: str
    task: str

class TaskResponse(BaseModel):
    task: TaskAssignment | None = None

class NextTaskResponse(BaseModel):
    continue_: bool = Field(..., alias="continue")
    task: TaskAssignment | None = None


from typing import Callable, Dict, Any, get_type_hints
import inspect

GREEN = "\033[92m"
RESET = "\033[0m"
BOLD = "\033[1m"
BOX_LINE = "-" * 110


def format_final_response(response: str) -> str:
    """Formats the final response to make it stand out in the terminal."""
    return f"\n{GREEN}{BOX_LINE}\n{BOLD}{response}{RESET}\n{GREEN}{BOX_LINE}{RESET}\n"


def python_type_to_json_schema(py_type: type) -> Dict[str, Any]:
    """Convert a Python type to a JSON Schema snippet."""
    try:
        mapping = {
            int: {"type": "integer"},
            float: {"type": "number"},
            str: {"type": "string"},
            bool: {"type": "boolean"},
        }

        if py_type in mapping:
            return mapping[py_type]

        print(f"Warning: Unsupported type '{py_type}'. Defaulting to 'string'.")
        return {"type": "string"}

    except Exception as e:
        print(f"Error converting Python type '{py_type}' to JSON Schema: {e}")
        return {"type": "string"}

def get_function_parameters(func: Callable) -> Dict[str, Any]:
    """
    Extracts parameters from a function and returns a JSON Schema.

    Returns a schema like:
    {
        "type": "object",
        "properties": {
            "a": {"type": "integer"},
            "b": {"type": "integer"}
        },
        "required": ["a", "b"]
    }
    """
    try:
        signature = inspect.signature(func)
        properties = {}
        required = []

        for name, param in signature.parameters.items():
            try:
                if param.annotation is inspect._empty:
                    json_type = {"type": "string"}
                else:
                    json_type = python_type_to_json_schema(param.annotation)

                properties[name] = json_type

                if param.default is inspect._empty:
                    required.append(name)

            except Exception as e:
                print(f"Error processing parameter '{name}' in function '{func.__name__}': {e}")
                properties[name] = {"type": "string"}  

        schema = {"type": "object", "properties": properties}

        if required:
            schema["required"] = required

        return schema

    except Exception as e:
        print(f"Error extracting function signature for '{func}': {e}")
        return {"type": "object", "properties": {}} 


from gentleagents.agents.agent import Agent
from gentleagents.tools.tool import add_numbers, sub_numbers, send_email
def test_agent():
    agent = Agent(name="Assistant",
                role="""You are an assistant that performs any tasks 
                if they can be done by the tools provided to you. You make sure a task is done if a tool
                for it is available""",
                tools={"add_numbers": add_numbers,"sub_numbers": sub_numbers, "send_email":send_email},
                model = "GROQ:llama-3.3-70b-versatile")
                # model = "GROQ:llama-3.3-70b-specdec")
                # model = "OPENAI:gpt-4-turbo")

    
    
    
    response = agent.start_agent("Can you add 5 and 7? And also subtract 9 from 20 thanks man, how are you btw?")
    # response = agent.start_agent("how are you btw?")
    # response = agent.start_agent("Can you subtract 5 from 7?")
    # response = agent.start_agent("Can you add 5 and 7? And also subtract 9 from 20 thanks man also email hi to abdullah_nizami@live.com")

    # print(response)

test_agent()

#TO DO: REMOVE COMMENTS, ADD PROPER HANDLING FOR WHEN WE CHANGE ENGINES.
#Run this test using python -m gentleagents.agents.test_agent at the project root
#root is at /GentleAgent



from gentleagents.engines.openai_client import interact

response = interact(
    system_prompt="You are a helpful assistant.",
    user_message="What is the capital of France?"
)

assistant_reply = response.choices[0].message.content

print(response)
print(assistant_reply)

#TO DO: REMOVE COMMENTS
#Run this file using python -m gentleagents.engines.test_openai at the project root.
#Root means at /GentleAgents


import json
from gentleagents.engines.openai_client import interact
from gentleagents.validations.models import TaskResponse, NextTaskResponse
from gentleagents.queries.query import get_team_system, get_team_followup, get_team_summary, get_team_initial_task

class Team:
    def __init__(self, lead_agent, agents):
        """
        Initialize a Team with a lead agent and multiple agents.
        
        Args:
            lead_agent (Agent): The agent responsible for managing tasks.
            agents (list): A list of other Agent instances available for task execution.
        """
        if lead_agent.tools:
            raise ValueError("Lead agent should not have any tools.")
        
        self.lead_agent = lead_agent
        self.agents = {agent.name: agent for agent in agents}
    
    def assign_tasks(self, task_prompt):
        try:
            self.lead_agent.chat_history = [{"role": "user", "content": task_prompt}]
            system_prompt = get_team_system(self.lead_agent.name)
            for agent in self.agents.values():
                system_prompt += f"- {agent.name}: {agent.role}. Tools: {', '.join(agent.tools.keys()) if agent.tools else 'None'}\n"
        
            system_prompt += get_team_initial_task()
            response = interact(system_prompt=system_prompt, user_message=task_prompt, model=self.lead_agent.model)
            if not response:
                return "Error: Could not assign tasks."

            response_text = response.choices[0].message.content.strip()
            if not response_text:
                return "Error: Lead agent returned an empty response. Ensure your prompt forces a JSON output."

            try:
                task_data = json.loads(response_text)
                parsed_response = TaskResponse(**task_data)
            except (json.JSONDecodeError, ValueError) as e:
                return f"Error parsing task response: {e}\nRaw response: {response_text}"

            all_agent_responses = []

            while parsed_response.task:
                assignment = parsed_response.task
                if assignment.agent in self.agents:
                    result = self.agents[assignment.agent].start_agent(assignment.task)
                    all_agent_responses.append(f"{assignment.agent} completed: {result}")
                else:
                    all_agent_responses.append(f"Error: No agent named {assignment.agent}.")

                self.lead_agent.chat_history.append({"role": "assistant", "content": "\n".join(all_agent_responses)})
                followup_prompt = get_team_followup(self.lead_agent.chat_history, self.lead_agent.name)
                followup_response = interact(system_prompt=f"{self.lead_agent.name} checking for further tasks",
                                            user_message=followup_prompt,
                                            model=self.lead_agent.model)
                if not followup_response:
                    break

                followup_text = followup_response.choices[0].message.content.strip()

                try:
                    next_task_data = json.loads(followup_text)
                    next_response = NextTaskResponse(**next_task_data)
                except (json.JSONDecodeError, ValueError) as e:
                    return f"Error parsing followup task response: {e}\nRaw response: {followup_text}"

                if not next_response.continue_ or not next_response.task:
                    break

                parsed_response = TaskResponse(task=next_response.task)

            summary_prompt = get_team_summary(task_prompt, all_agent_responses)


            final_response = interact(system_prompt=f"{self.lead_agent.name} summarizing tasks",
                                    user_message=summary_prompt,
                                    model=self.lead_agent.model)
            return final_response.choices[0].message.content if final_response else "Error in final summary."

        except Exception as e:
            return f"Critical error in task assignment: {e}"




from gentleagents.agents.agent import Agent
from gentleagents.teams.team import Team
from gentleagents.tools.tool import add_numbers, sub_numbers, send_email

def test():
    add_agent = Agent(
        name="Addition Agent",
        role="Performs simple additions using the provided tool.",
        tools={"add_numbers": add_numbers},
        model="GROQ:llama-3.3-70b-versatile"
    )

    sub_agent = Agent(
        name="Subtraction Agent",
        role="Performs simple subtractions using the provided tool.",
        tools={"sub_numbers": sub_numbers},
        model="GROQ:llama-3.3-70b-versatile"
    )

    email_agent = Agent(
        name="Email Agent",
        role="Sends emails using the provided tool.",
        tools={"send_email": send_email},
        model="GROQ:llama-3.3-70b-versatile"
    )

    lead_agent = Agent(
        name="Lead Agent",
        role="Manages task delegation among specialized agents and summarizes results.",
        tools={},
        model="GROQ:llama-3.3-70b-versatile"
    )

    team = Team(lead_agent, [add_agent, sub_agent, email_agent])

    # tasks = "Add 5 and 3. And email the answer to abdullah_nizami@live.com, and oh also subtract 4 from 10 thanks!"
    tasks = "Add 5 and 3. and oh also subtract 4 from 10 thanks!"

    result = team.assign_tasks(tasks)

    print("\nFinal Team Output:\n", result)


test()



import openai
import time
import sys
from gentleagents.loadenv.loadenv import OPENAI_API_KEY, GROQ_API_KEY, GROQ_API_URL


MAX_RETRIES = 3  
RETRY_DELAY = 2  
RED = "\033[91m" 
RESET = "\033[0m"  

def interact(system_prompt, user_message, tools=None, model=None):
    """
    Sends a message to OpenAI's GPT model and returns the response.

    Args:
        system_prompt (str): The system-level instruction to guide the model.
        user_message (str): The user's input message.
        tools (list, optional): List of tool definitions for function calling.
        model (str, optional): OpenAI model to use (default is GPT-4-turbo).

    Returns:
        dict: The response from OpenAI API, or None if an error occurs.
    """


    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": user_message}
    ]
    
    provider = model.split(":")[0] if ":" in model else model
    model = model.split(":")[1] if ":" in model else model
    
    if provider == "GROQ":
        if not GROQ_API_KEY or not GROQ_API_URL:
            sys.stderr.write(f"{RED} ERROR: GROQ_API_KEY or GROQ_API_URL is not set. Please configure your API key and the Groq API URL.{RESET}\n")
            sys.exit(1)  
        client = openai.OpenAI(base_url=GROQ_API_URL, api_key=GROQ_API_KEY)

    
    elif provider == "OPENAI":
        if not OPENAI_API_KEY:
            sys.stderr.write(f"{RED} ERROR: OPENAI_API_KEY is not set. Please configure your API key.{RESET}\n")
            sys.exit(1)  
            raise ValueError("")
        openai.api_key = OPENAI_API_KEY
        client = openai.OpenAI()  

    params = {
        "model": model,
        "messages": messages,
    }

    if tools:
        params["tools"] = tools  
        params["tool_choice"] = "auto"


    for attempt in range(1, MAX_RETRIES + 1):
        try:
            response = client.chat.completions.create(**params)
            return response  
        
        except openai.APIError as e:
            print(f"OpenAI API error (attempt {attempt}): {e}")
        except openai.RateLimitError as e:
            print(f"Rate limit exceeded (attempt {attempt}): {e}")
        except openai.OpenAIError as e:
            print(f"Unexpected OpenAI error (attempt {attempt}): {e}")
        except Exception as e:
            print(f"Unknown error (attempt {attempt}): {e}")

        if attempt < MAX_RETRIES:
            time.sleep(RETRY_DELAY)  

    print("Max retries reached. Failed to get a response.")
    return None  



from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from gentleagents.loadenv.loadenv import SENDER_EMAIL, SENDER_PASSWORD
import smtplib

def add_numbers(a: int, b: int) -> int:
    """Use this function to add two numbers.
    
    Args:
        a: int: The first number
        b: int: The second number

    Returns:
        int: sum of a + b
    """
    return a + b

def sub_numbers(a: int, b: int) -> int:
    """Use this function to subtract two numbers.
    
    Args:
        a: int: The first number
        b: int: The second number

    Returns:
        int: result of a - b
    """
    return a - b

def send_email(recipient: str, subject: str, content: str) -> str: 
    """Use this function to send an email.

    Args:
        recipient:str: the recipients' email, comma seperated if multiple
        subject:str: the email subject
        content:str: the email body
    """
    
    sender_email = SENDER_EMAIL
    sender_password = SENDER_PASSWORD

    message = MIMEMultipart()
    message["From"] = sender_email
    message["To"] = recipient
    message["Subject"] = subject

    message.attach(MIMEText(content, "plain"))

    try:
        with smtplib.SMTP_SSL("smtp.gmail.com", 465) as server:
            server.login(sender_email, sender_password)
            server.sendmail(sender_email, recipient, message.as_string())
        
        status = f"Successfully sent email to {recipient}"
        return status
    except Exception as e:
        print(f"Error in Email tool: {str(e)}")
        status = "Failed to send email."
        return status


from dotenv import load_dotenv
import os

DEFAULT_ENV_PATH = os.path.join(os.getcwd(), ".env")

ENV_PATH = os.getenv("GENTLEAGENTS_ENV_PATH", DEFAULT_ENV_PATH)

load_dotenv(ENV_PATH)

OPENAI_API_KEY = os.getenv('OPENAI_API_KEY')
SENDER_EMAIL = os.getenv('SENDER_EMAIL')
SENDER_PASSWORD = os.getenv('SENDER_PASSWORD')
GROQ_API_KEY = os.getenv('GROQ_API_KEY')
GROQ_API_URL = os.getenv('GROQ_API_URL')



def get_agent_summary(chat_history):
    return (
        "Based on the conversation so far and the actions taken (including the tool outputs below),"
        " please provide what steps you took, which tools you used, and include a polite, short, concise general answer"
        " IF my query contains something that wasn't processed by the tools..\n\n"
        f"Conversation history:\n{chat_history}\n\n"
        "Final answer:"
    )

def get_team_system(lead_agent_name):
    return (
        f"You are {lead_agent_name}, the lead agent. Your job is to distribute tasks among specialized agents.\n\n"
        "You do not judge the responses by the agents, nor do you suggest any changes to the tools. You are simply a messenger.\n"
        "Here are the available agents and their tools:\n"
)

def get_team_followup(chat_history, lead_agent_name):
    return (
        f"You are {lead_agent_name}, the lead agent in this task delegation.\n"
        "Based on the conversation so far and the outcomes of the executed tasks, determine if there is another task required.\n"
        "From the conversation history, make SURE all tasks have been done!\n"
        "Return a JSON response in the following format and NOTHING ELSE:\n"
        '{ "continue": <true/false>, "task": {"agent": "<Agent Name>", "task": "<Task Description>"} }\n'
        "If no further tasks are needed, set \"continue\" to false and return null for \"task\".\n\n"
        "Conversation history:\n"
        f"{chat_history}\n"
    )


def get_team_summary(task_prompt, all_agent_responses):
    return (
        "You are the lead agent in this task delegation.\n"
        "Provide a factual summary of the task executions without any commentary, suggestions, or judgment. For each task, list:\n"
        " - The task assigned.\n"
        " - The agent that executed the task.\n"
        " - The tool used.\n"
        " - The exact output returned by the tool.\n\n"
        "Return ONLY the facts in the exact format above.\n\n"
        f"Task Prompt: {task_prompt}\n\n"
        "Agent Responses:\n"
        "\n".join(all_agent_responses)
    )

def get_team_initial_task():
    return (
        "\nAnalyze the task prompt and decide which agent should handle the VERY FIRST task.\n"
        "Return a JSON response with this format and NOTHING ELSE:\n"
        '{ "task": {"agent": "<Agent Name>", "task": "<Task Description>"} }'
    )
